from pathlib import Path

from git.repo import Repo
from tqdm import tqdm

from seagoat.cache import Cache
from seagoat.common import IGNORED_BRANCHES
from seagoat.common import SUPPORTED_FILE_TYPES
from seagoat.file import File


class Repository:
    def __init__(self, path: str, cache: Cache) -> None:
        self._repo = Repo(path)
        self._cache = cache
        self.path = path

    def _get_all_commits(self):
        for branch in tqdm(self._repo.branches, desc="Analyzing branches"):
            if branch.name in IGNORED_BRANCHES:
                continue
            if (
                self._cache.data["last_analyzed_version_of_branch"].get(branch.name)
                == branch.commit.hexsha
            ):
                continue
            for commit in self._repo.iter_commits(branch):
                self._cache.data["required_commits"].add(commit.hexsha)
            self._cache.data["last_analyzed_version_of_branch"][
                branch.name
            ] = branch.commit.hexsha

        return (
            self._repo.commit(commit) for commit in self._cache.data["required_commits"]
        )

    def analyze_files(self):
        for commit in tqdm(
            self._get_all_commits(),
            desc="Analyzing commits",
            total=len(self._cache.data["required_commits"]),
        ):
            if commit.hexsha in self._cache.data["commits_already_analyzed"]:
                continue
            self._cache.data["commits_already_analyzed"].add(commit.hexsha)
            for path in commit.stats.files:  # type: ignore[reportGeneralTypeIssues]
                if Path(path).suffix not in SUPPORTED_FILE_TYPES:
                    continue
                if path not in self._cache.data["file_data"]:
                    self._cache.data["file_data"][path] = File(
                        path, Path(self._repo.working_dir) / path
                    )
                self._cache.data["file_data"][path].add_commit(commit)

        self._sort_files()
        self._cache.persist()

    def _sort_files(self):
        self._cache.data["sorted_files"] = list(
            sorted(
                self._cache.data["file_data"].keys(),
                key=lambda x: self._cache.data["file_data"][x].get_score(),
                reverse=True,
            )
        )

    def top_files(self):
        return [
            self._cache.data["file_data"][path]
            for path in self._cache.data["sorted_files"]
        ]

    def get_file(self, path: str):
        if path not in self._cache.data["file_data"]:
            raise RuntimeError("File not found or not analyzed yet")
        return self._cache.data["file_data"][path]



import os
import sys
from functools import cache
from itertools import islice

import click
import requests
from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import get_lexer_for_filename
from pygments.lexers.javascript import JavascriptLexer
from pygments.lexers.javascript import TypeScriptLexer

from seagoat import __version__
from seagoat.server import get_server_info_file
from seagoat.server import load_server_info


# pylint: disable-next=too-few-public-methods
class ExitCode:
    SERVER_NOT_RUNNING = 3


def query_server(query, server_address, repo_path):
    try:
        response = requests.get(f"{server_address}/query/{query}")
        response.raise_for_status()
    except (requests.exceptions.ConnectionError, requests.exceptions.RequestException):
        print(
            f"The SeaGOAT server is not running. "
            f"Please start the server using the following command:\n\n"
            f"seagoat-server start {repo_path}\n"
        )
        sys.exit(ExitCode.SERVER_NOT_RUNNING)

    return response.json()["results"]


@cache
def get_highlighted_lines(file_name: str):
    with open(file_name, "r", encoding="utf-8") as source_code_file:
        code = source_code_file.read()

    if file_name.endswith(".md"):
        return code.splitlines()

    if file_name.endswith(".jsx"):
        lexer = JavascriptLexer()
    elif file_name.endswith(".tsx"):
        lexer = TypeScriptLexer()
    else:
        lexer = get_lexer_for_filename(file_name)

    result = highlight(code, lexer, TerminalFormatter())

    return result.splitlines()


def print_result_line(result, line, color_enabled):
    if color_enabled:
        highlighted_lines = get_highlighted_lines(str(result["full_path"]))
        print(
            f"{result['path']}:{click.style(str(line), bold=True)}:{highlighted_lines[line - 1]}"
        )
    else:
        for line_content in result["lines"]:
            if line_content["line"] == line:
                print(f"{result['path']}:{line}:{line_content['line_text']}")
                break


def iterate_result_lines(results):
    for result in results:
        for line in result.get("lines", []):
            yield result, line


@click.command()
@click.argument("query")
@click.argument("repo_path", required=False, default=os.getcwd())
@click.option(
    "--no-color",
    is_flag=True,
    help="Disable formatting. Automatically enabled when part of a bash pipeline.",
)
@click.option(
    "-l",
    "--max-results",
    type=int,
    default=None,
    help="Limit the number of result lines",
)
@click.version_option(version=__version__, prog_name="seagoat")
def seagoat(query, repo_path, no_color, max_results):
    """
    Query your codebase for your QUERY in the Git repository REPO_PATH.
    Your query can contain keywords, regular expression patterns,
    or a description of what you are looking for.

    When REPO_PATH is not specified, the current working directory is
    assumed to be the repository path.

    In order to use seagoat in your repository, you need to run a server
    that will analyze your codebase. Check seagoat-server --help for more details.
    """
    server_info_file = get_server_info_file(repo_path)
    _, __, ___, server_address = load_server_info(server_info_file)
    results = query_server(query, server_address, repo_path)

    color_enabled = os.isatty(0) and not no_color

    for result, result_line in islice(iterate_result_lines(results), max_results):
        print_result_line(result, result_line["line"], color_enabled)


if __name__ == "__main__":
    # pylint: disable-next=no-value-for-parameter
    seagoat()



# pylint: disable=too-few-public-methods
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Set


@dataclass(frozen=True)
class ResultLine:
    line: int
    vector_distance: float
    line_text: str

    def _get_number_of_exact_matches(self, query: str) -> int:
        terms = re.split(r"\s+", query)
        pattern = ".*".join(map(re.escape, terms))

        if re.search(pattern, self.line_text, re.IGNORECASE):
            return 1
        return 0

    def get_score(self, query: str) -> float:
        return self.vector_distance / (1 + self._get_number_of_exact_matches(query))

    def to_json(self):
        return {
            "line": self.line,
            "line_text": self.line_text,
        }


class Result:
    def __init__(self, path: str, full_path: Path) -> None:
        self.path: str = path
        self.full_path: Path = full_path
        self.lines: Set[ResultLine] = set()
        self.line_texts = self._read_lines()

    def extend(self, other) -> None:
        self.lines.update(other.lines)

    def _read_lines(self):
        with open(self.full_path, encoding="utf-8") as source_code_file:
            return source_code_file.read().splitlines()

    def add_line(self, line: int, vector_distance: float) -> None:
        self.lines.add(ResultLine(line, vector_distance, self.line_texts[line - 1]))

    def get_best_score(self, query: str) -> float:
        return min(self.lines, key=lambda item: item.get_score(query)).get_score(query)

    def get_lines(self, query: str):
        best_score = self.get_best_score(query)

        return list(
            sorted(
                set(
                    result_line.line
                    for result_line in self.lines
                    if result_line.get_score(query) <= best_score * 1.2
                )
            )
        )

    def to_json(self):
        return {
            "path": self.path,
            "full_path": str(self.full_path),
            "lines": [
                line.to_json()
                for line in sorted(self.lines, key=lambda item: item.line)
            ],
        }
